#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <richedit.h>
#include "ConvFunc.h"
#include "StackFunc.h"
#include "StrFunc.h"

#define AEC_FUNCTIONS
#include "AkelEdit.h"
#include "AkelDLL.h"


//Include conversion functions
#define xatoiA
#define xatoiW
#include "ConvFunc.h"

//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackClear
#include "StackFunc.h"

//Include string functions
#define WideCharUpper
#define xstrcmpiA
#define xstrcmpiW
#define xstrcmpinA
#define xstrcmpinW
#include "StrFunc.h"

//Defines
#define STRID_WORDWRAPOFF    1

//LinesOrder action
#define LO_SORTSTRASC        1
#define LO_SORTSTRDESC       2
#define LO_SORTINTASC        3
#define LO_SORTINTDESC       4
#define LO_FIXWRAP           5
#define LO_REVERSE           6
#define LO_GETDUPLICATES     7
#define LO_REMOVEDUPLICATES  8

typedef struct _LINEITEMARRAYA {
  char *pLine;
  int nLineLen;
  int nOffset;
} LINEITEMARRAYA;

typedef struct _LINEITEMARRAYW {
  wchar_t *wpLine;
  int nLineLen;
  int nOffset;
} LINEITEMARRAYW;

typedef struct _LINEITEMSTACKA {
  struct _LINEITEMSTACKA *next;
  struct _LINEITEMSTACKA *prev;
  char *pLine;
  int nLineLen;
} LINEITEMSTACKA;

typedef struct _LINEITEMSTACKW {
  struct _LINEITEMSTACKW *next;
  struct _LINEITEMSTACKW *prev;
  wchar_t *wpLine;
  int nLineLen;
} LINEITEMSTACKW;

typedef struct _DUPLICATEITEMA {
  struct _DUPLICATEITEMA *next;
  struct _DUPLICATEITEMA *prev;
  char *pLine;
  int nLineLen;
  char *pLineStr;
  int nLineStrLen;
} DUPLICATEITEMA;

typedef struct _DUPLICATEITEMW {
  struct _DUPLICATEITEMW *next;
  struct _DUPLICATEITEMW *prev;
  wchar_t *wpLine;
  int nLineLen;
  wchar_t *wpLineStr;
  int nLineStrLen;
} DUPLICATEITEMW;

//Functions prototypes
void LinesOrderA(HWND hWnd, HSTACK *hStack, int nOrder);
void LinesOrderW(HWND hWnd, HSTACK *hStack, int nOrder);
int ReplaceSelFromStackA(HWND hWnd, HSTACK *hStack);
int ReplaceSelFromStackW(HWND hWnd, HSTACK *hStack);
DWORD CALLBACK ReplaceSelFromStackCallbackA(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb);
DWORD CALLBACK ReplaceSelFromStackCallbackW(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb);
DUPLICATEITEMA* StackPushRemoveDuplicatesA(HSTACK *hStack, char *pLine, int nLineLen, char *pLineStr, int nLineStrLen);
DUPLICATEITEMW* StackPushRemoveDuplicatesW(HSTACK *hStack, wchar_t *wpLine, int nLineLen, wchar_t *wpLineStr, int nLineStrLen);
LINEITEMSTACKA* StackInsertLineA(HSTACK *hStack, char *pLine, int nLineLen, int nIndex);
LINEITEMSTACKW* StackInsertLineW(HSTACK *hStack, wchar_t *wpLine, int nLineLen, int nIndex);
void StackLinesFreeA(HSTACK *hStack);
void StackLinesFreeW(HSTACK *hStack);
void StackFree(HSTACK *hStack);
int CompareLineStrA(const void *elem1, const void *elem2);
int CompareLineStrW(const void *elem1, const void *elem2);
int CompareLineIntA(const void *elem1, const void *elem2);
int CompareLineIntW(const void *elem1, const void *elem2);
void swap_item(register unsigned char *src, register unsigned char *dest, register unsigned int size);
void quicksort(void *base, unsigned int count, unsigned int width, int (*compare)(const void *, const void *));
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
void InitCommon(PLUGINDATA *pd);

//Global variables
HWND hMainWnd;
HWND hWndEdit;
BOOL bOldWindows;
BOOL bAkelEdit;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
HSTACK hLinesStack={0};
int nLineOffset=0;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="Lines";
}

//Plugin extern function
extern "C" void __declspec(dllexport) SortStrAsc(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_SORTSTRASC);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_SORTSTRASC);
}

extern "C" void __declspec(dllexport) SortStrDesc(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_SORTSTRDESC);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_SORTSTRDESC);
}

extern "C" void __declspec(dllexport) SortIntAsc(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_SORTINTASC);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_SORTINTASC);
}

extern "C" void __declspec(dllexport) SortIntDesc(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_SORTINTDESC);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_SORTINTDESC);
}

extern "C" void __declspec(dllexport) FixWrap(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_FIXWRAP);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_FIXWRAP);
}

extern "C" void __declspec(dllexport) Reverse(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_REVERSE);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_REVERSE);
}

extern "C" void __declspec(dllexport) GetDuplicates(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_GETDUPLICATES);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_GETDUPLICATES);
}

extern "C" void __declspec(dllexport) RemoveDuplicates(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  InitCommon(pd);

  if (bOldWindows)
    LinesOrderA(hWndEdit, &hLinesStack, LO_REMOVEDUPLICATES);
  else
    LinesOrderW(hWndEdit, &hLinesStack, LO_REMOVEDUPLICATES);
}

void LinesOrderA(HWND hWnd, HSTACK *hStack, int nOrder)
{
  AESELECTION aesInitial;
  AECHARINDEX ciInitialCaret;
  AECHARINDEX ciCount;
  EDITINFO ei;
  GETTEXTRANGE gtr;
  CHARRANGE cr;
  HSTACK hTempStack={0};
  LINEITEMARRAYA *lpLinesArray=NULL;
  char *pLine;
  char *pText;
  char *pLineStr;
  int nMinLine;
  int nMaxLine;
  int nMinLineIndex;
  int nMaxLineIndex;
  int nMaxLineLength;
  int nLineLen;
  int nLineStrLen;
  int nLineIndex;
  int nLinesCount=0;
  int nTextLen;
  int nTextCount;
  int i;
  BOOL bColumnSel=FALSE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
  {
    if (!ei.bReadOnly)
    {
      if (bAkelEdit)
      {
        if (SendMessage(hWnd, AEM_GETSEL, (WPARAM)&ciInitialCaret, (LPARAM)&aesInitial))
          bColumnSel=(aesInitial.dwFlags & AESELT_COLUMNON);
      }

      //Select full lines
      SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);
      if (cr.cpMin == cr.cpMax)
      {
        i=SendMessage(hMainWnd, AKD_GETTEXTLENGTH, (WPARAM)hWnd, 0);
        cr.cpMin=0;
        cr.cpMax=i;
      }
      else
      {
        //Block sorting
        bColumnSel=SendMessage(hWnd, AEM_GETCOLUMNSEL, 0, 0);
      }
      nMinLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, cr.cpMin);
      nMaxLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, cr.cpMax);
      nMinLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMinLine, 0);
      nMaxLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMaxLine, 0);
      nMaxLineLength=SendMessage(hWnd, EM_LINELENGTH, cr.cpMax, 0);

      if (nMaxLineIndex == cr.cpMax) --nMaxLine;
      else if (nMaxLineLength) cr.cpMax=nMaxLineIndex + nMaxLineLength + 1;
      cr.cpMin=nMinLineIndex;

      if (nMinLine != nMaxLine)
      {
        if (nOrder == LO_FIXWRAP)
        {
          if (!ei.bWordWrap)
          {
            MessageBoxA(hMainWnd, GetLangStringA(wLangSystem, STRID_WORDWRAPOFF), "Lines plugin", MB_OK|MB_ICONEXCLAMATION);
            return;
          }

          for (i=nMinLine; i <= nMaxLine; ++i)
          {
            nLineIndex=SendMessage(hWnd, EM_LINEINDEX, i, 0);
            nLineLen=SendMessage(hWnd, EM_LINELENGTH, nLineIndex, 0);

            if (pLine=(char *)GlobalAlloc(GPTR, nLineLen + 2))
            {
              *(WORD *)pLine=(WORD)nLineLen;
              nLineLen=SendMessageA(hWnd, EM_GETLINE, i, (LPARAM)pLine);
              pLine[nLineLen]='\0';

              StackInsertLineA(hStack, pLine, nLineLen, -1);
            }
            else break;
          }
        }
        else
        {
          gtr.cpMin=cr.cpMin;
          gtr.cpMax=cr.cpMax;

          if (nTextLen=SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr))
          {
            //Create empty array for quicksort
            if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
            {
              pText=(char *)gtr.pText;
              nLinesCount=1;

              for (nTextCount=0; nTextCount < nTextLen; ++nTextCount)
              {
                if (pText[nTextCount] == '\r')
                  ++nLinesCount;
              }
              lpLinesArray=(LINEITEMARRAYA *)GlobalAlloc(GPTR, nLinesCount * sizeof(LINEITEMARRAYA));
              nLinesCount=0;
            }

            if (bColumnSel) SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTSELCHAR, (LPARAM)&ciCount);
            pLine=(char *)gtr.pText;
            nLineLen=0;

            for (nTextCount=0; nTextCount <= nTextLen; ++nTextCount)
            {
              if (pLine[nLineLen] == '\r' || nTextCount == nTextLen)
              {
                if (nTextCount == nTextLen && nLineLen == 0)
                  break;
                pLine[nLineLen]='\0';

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                    nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
                {
                  if (bColumnSel)
                  {
                    lpLinesArray[nLinesCount].pLine=pLine;
                    lpLinesArray[nLinesCount].nLineLen=nLineLen;
                    lpLinesArray[nLinesCount].nOffset=min(ciCount.lpLine->nSelStart, nLineLen);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&ciCount);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_NEXTLINE, (LPARAM)&ciCount);
                  }
                  else
                  {
                    lpLinesArray[nLinesCount].pLine=pLine;
                    lpLinesArray[nLinesCount].nLineLen=nLineLen;
                    lpLinesArray[nLinesCount].nOffset=0;
                  }
                  nLinesCount++;
                }
                else if (nOrder == LO_REVERSE)
                {
                  StackInsertLineA(hStack, pLine, nLineLen, 1);
                }
                else if (nOrder == LO_GETDUPLICATES ||
                         nOrder == LO_REMOVEDUPLICATES)
                {
                  if (bColumnSel)
                  {
                    if (nLineStrLen=min(ciCount.lpLine->nSelEnd, nLineLen) - min(ciCount.lpLine->nSelStart, nLineLen))
                      pLineStr=pLine + min(ciCount.lpLine->nSelStart, nLineLen);
                    else
                    {
                      pLineStr=pLine;
                      nLineStrLen=nLineLen;
                    }
                    if (nOrder == LO_GETDUPLICATES)
                    {
                      if (!StackPushRemoveDuplicatesA(&hTempStack, pLine, nLineLen, pLineStr, nLineStrLen))
                        StackInsertLineA(hStack, pLine, nLineLen, -1);
                    }
                    else StackPushRemoveDuplicatesA(hStack, pLine, nLineLen, pLineStr, nLineStrLen);

                    SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&ciCount);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_NEXTLINE, (LPARAM)&ciCount);
                  }
                  else
                  {
                    if (nOrder == LO_GETDUPLICATES)
                    {
                      if (!StackPushRemoveDuplicatesA(&hTempStack, pLine, nLineLen, pLine, nLineLen))
                        StackInsertLineA(hStack, pLine, nLineLen, -1);
                    }
                    else StackPushRemoveDuplicatesA(hStack, pLine, nLineLen, pLine, nLineLen);
                  }
                }
                pLine+=nLineLen + 1;
                nLineLen=0;
              }
              else ++nLineLen;
            }

            //Sort and fill the stack
            if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
            {
              if (lpLinesArray)
              {
                int nUpDown=0;

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTINTASC)
                  nUpDown=-1;
                else if (nOrder == LO_SORTSTRDESC || nOrder == LO_SORTINTDESC)
                  nUpDown=1;

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC)
                  quicksort(lpLinesArray, nLinesCount, sizeof(LINEITEMARRAYA), CompareLineStrA);
                else if (nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
                  quicksort(lpLinesArray, nLinesCount, sizeof(LINEITEMARRAYA), CompareLineIntA);

                for (i=0; i < nLinesCount; ++i)
                {
                  StackInsertLineA(hStack, lpLinesArray[i].pLine, lpLinesArray[i].nLineLen, nUpDown);
                }
                GlobalFree((HGLOBAL)lpLinesArray);
              }
            }
            else if (nOrder == LO_GETDUPLICATES)
            {
              StackFree(&hTempStack);
            }
          }
        }
        SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
        i=ReplaceSelFromStackA(hWnd, hStack);
        cr.cpMax=cr.cpMin + i / sizeof(wchar_t);

        if (bAkelEdit && bColumnSel)
        {
          SendMessage(hWnd, AEM_INDEXUPDATE, 0, (LPARAM)&aesInitial.crSel.ciMin);
          SendMessage(hWnd, AEM_INDEXUPDATE, 0, (LPARAM)&aesInitial.crSel.ciMax);

          if (!AEC_IndexCompare(&aesInitial.crSel.ciMin, &ciInitialCaret))
            SendMessage(hWnd, AEM_SETSEL, (WPARAM)&aesInitial.crSel.ciMin, (LPARAM)&aesInitial);
          else
            SendMessage(hWnd, AEM_SETSEL, (WPARAM)&aesInitial.crSel.ciMax, (LPARAM)&aesInitial);
        }
        else SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);

        //Free memory
        if (nOrder == LO_FIXWRAP)
        {
          StackLinesFreeA(hStack);
        }
        else
        {
          StackFree(hStack);
          SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
        }
      }
    }
    else MessageBeep(MB_OK);
  }
}

void LinesOrderW(HWND hWnd, HSTACK *hStack, int nOrder)
{
  AESELECTION aesInitial;
  AECHARINDEX ciInitialCaret;
  AECHARINDEX ciCount;
  EDITINFO ei;
  GETTEXTRANGE gtr;
  CHARRANGE cr;
  HSTACK hTempStack={0};
  LINEITEMARRAYW *lpLinesArray=NULL;
  wchar_t *wpText;
  wchar_t *wpLine;
  wchar_t *wpLineStr;
  int nMinLine;
  int nMaxLine;
  int nMinLineIndex;
  int nMaxLineIndex;
  int nMaxLineLength;
  int nLineLen;
  int nLineStrLen;
  int nLineIndex;
  int nLinesCount=0;
  int nTextLen;
  int nTextCount;
  int i;
  BOOL bColumnSel=FALSE;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)hWnd, (LPARAM)&ei))
  {
    if (!ei.bReadOnly)
    {
      if (bAkelEdit)
      {
        if (SendMessage(hWnd, AEM_GETSEL, (WPARAM)&ciInitialCaret, (LPARAM)&aesInitial))
          bColumnSel=(aesInitial.dwFlags & AESELT_COLUMNON);
      }

      //Select full lines
      SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);
      if (cr.cpMin == cr.cpMax)
      {
        i=SendMessage(hMainWnd, AKD_GETTEXTLENGTH, (WPARAM)hWnd, 0);
        cr.cpMin=0;
        cr.cpMax=i;
      }
      else
      {
        //Block sorting
        bColumnSel=SendMessage(hWnd, AEM_GETCOLUMNSEL, 0, 0);
      }
      nMinLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, cr.cpMin);
      nMaxLine=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, cr.cpMax);
      nMinLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMinLine, 0);
      nMaxLineIndex=SendMessage(hWnd, EM_LINEINDEX, nMaxLine, 0);
      nMaxLineLength=SendMessage(hWnd, EM_LINELENGTH, cr.cpMax, 0);

      if (nMaxLineIndex == cr.cpMax) --nMaxLine;
      else if (nMaxLineLength) cr.cpMax=nMaxLineIndex + nMaxLineLength + 1;
      cr.cpMin=nMinLineIndex;

      if (nMinLine != nMaxLine)
      {
        if (nOrder == LO_FIXWRAP)
        {
          if (!ei.bWordWrap)
          {
            MessageBoxW(hMainWnd, GetLangStringW(wLangSystem, STRID_WORDWRAPOFF), L"Lines plugin", MB_OK|MB_ICONEXCLAMATION);
            return;
          }

          for (i=nMinLine; i <= nMaxLine; ++i)
          {
            nLineIndex=SendMessage(hWnd, EM_LINEINDEX, i, 0);
            nLineLen=SendMessage(hWnd, EM_LINELENGTH, nLineIndex, 0);

            if (wpLine=(wchar_t *)GlobalAlloc(GPTR, nLineLen * sizeof(wchar_t) + 2))
            {
              *(WORD *)wpLine=(WORD)nLineLen;
              nLineLen=SendMessageW(hWnd, EM_GETLINE, i, (LPARAM)wpLine);
              wpLine[nLineLen]='\0';

              StackInsertLineW(hStack, wpLine, nLineLen, -1);
            }
            else break;
          }
        }
        else
        {
          gtr.cpMin=cr.cpMin;
          gtr.cpMax=cr.cpMax;

          if (nTextLen=SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr))
          {
            //Create empty array for quicksort
            if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
            {
              wpText=(wchar_t *)gtr.pText;
              nLinesCount=1;

              for (nTextCount=0; nTextCount < nTextLen; ++nTextCount)
              {
                if (wpText[nTextCount] == '\r')
                  ++nLinesCount;
              }
              lpLinesArray=(LINEITEMARRAYW *)GlobalAlloc(GPTR, nLinesCount * sizeof(LINEITEMARRAYW));
              nLinesCount=0;
            }

            if (bColumnSel) SendMessage(hWnd, AEM_GETINDEX, AEGI_FIRSTSELCHAR, (LPARAM)&ciCount);
            wpLine=(wchar_t *)gtr.pText;
            nLineLen=0;

            for (nTextCount=0; nTextCount <= nTextLen; ++nTextCount)
            {
              if (wpLine[nLineLen] == '\r' || nTextCount == nTextLen)
              {
                if (nTextCount == nTextLen && nLineLen == 0)
                  break;
                wpLine[nLineLen]='\0';

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                    nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
                {
                  if (bColumnSel)
                  {
                    lpLinesArray[nLinesCount].wpLine=wpLine;
                    lpLinesArray[nLinesCount].nLineLen=nLineLen;
                    lpLinesArray[nLinesCount].nOffset=min(ciCount.lpLine->nSelStart, nLineLen);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&ciCount);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_NEXTLINE, (LPARAM)&ciCount);
                  }
                  else
                  {
                    lpLinesArray[nLinesCount].wpLine=wpLine;
                    lpLinesArray[nLinesCount].nLineLen=nLineLen;
                    lpLinesArray[nLinesCount].nOffset=0;
                  }
                  nLinesCount++;
                }
                else if (nOrder == LO_REVERSE)
                {
                  StackInsertLineW(hStack, wpLine, nLineLen, 1);
                }
                else if (nOrder == LO_GETDUPLICATES ||
                         nOrder == LO_REMOVEDUPLICATES)
                {
                  if (bColumnSel)
                  {
                    if (nLineStrLen=min(ciCount.lpLine->nSelEnd, nLineLen) - min(ciCount.lpLine->nSelStart, nLineLen))
                      wpLineStr=wpLine + min(ciCount.lpLine->nSelStart, nLineLen);
                    else
                    {
                      wpLineStr=wpLine;
                      nLineStrLen=nLineLen;
                    }
                    if (nOrder == LO_GETDUPLICATES)
                    {
                      if (!StackPushRemoveDuplicatesW(&hTempStack, wpLine, nLineLen, wpLineStr, nLineStrLen))
                        StackInsertLineW(hStack, wpLine, nLineLen, -1);
                    }
                    else StackPushRemoveDuplicatesW(hStack, wpLine, nLineLen, wpLineStr, nLineStrLen);

                    SendMessage(hWnd, AEM_GETINDEX, AEGI_WRAPLINEEND, (LPARAM)&ciCount);
                    SendMessage(hWnd, AEM_GETINDEX, AEGI_NEXTLINE, (LPARAM)&ciCount);
                  }
                  else
                  {
                    if (nOrder == LO_GETDUPLICATES)
                    {
                      if (!StackPushRemoveDuplicatesW(&hTempStack, wpLine, nLineLen, wpLine, nLineLen))
                        StackInsertLineW(hStack, wpLine, nLineLen, -1);
                    }
                    else StackPushRemoveDuplicatesW(hStack, wpLine, nLineLen, wpLine, nLineLen);
                  }
                }
                wpLine+=nLineLen + 1;
                nLineLen=0;
              }
              else ++nLineLen;
            }

            //Sort and fill the stack
            if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC ||
                nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
            {
              if (lpLinesArray)
              {
                int nUpDown=0;

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTINTASC)
                  nUpDown=-1;
                else if (nOrder == LO_SORTSTRDESC || nOrder == LO_SORTINTDESC)
                  nUpDown=1;

                if (nOrder == LO_SORTSTRASC || nOrder == LO_SORTSTRDESC)
                  quicksort(lpLinesArray, nLinesCount, sizeof(LINEITEMARRAYW), CompareLineStrW);
                else if (nOrder == LO_SORTINTASC || nOrder == LO_SORTINTDESC)
                  quicksort(lpLinesArray, nLinesCount, sizeof(LINEITEMARRAYW), CompareLineIntW);

                for (i=0; i < nLinesCount; ++i)
                {
                  StackInsertLineW(hStack, lpLinesArray[i].wpLine, lpLinesArray[i].nLineLen, nUpDown);
                }
                GlobalFree((HGLOBAL)lpLinesArray);
              }
            }
            else if (nOrder == LO_GETDUPLICATES)
            {
              StackFree(&hTempStack);
            }
          }
        }
        SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);
        i=ReplaceSelFromStackW(hWnd, hStack);
        cr.cpMax=cr.cpMin + i / sizeof(wchar_t);

        if (bAkelEdit && bColumnSel)
        {
          SendMessage(hWnd, AEM_INDEXUPDATE, 0, (LPARAM)&aesInitial.crSel.ciMin);
          SendMessage(hWnd, AEM_INDEXUPDATE, 0, (LPARAM)&aesInitial.crSel.ciMax);

          if (!AEC_IndexCompare(&aesInitial.crSel.ciMin, &ciInitialCaret))
            SendMessage(hWnd, AEM_SETSEL, (WPARAM)&aesInitial.crSel.ciMin, (LPARAM)&aesInitial);
          else
            SendMessage(hWnd, AEM_SETSEL, (WPARAM)&aesInitial.crSel.ciMax, (LPARAM)&aesInitial);
        }
        else SendMessage(hWnd, EM_EXSETSEL, 0, (LPARAM)&cr);

        //Free memory
        if (nOrder == LO_FIXWRAP)
        {
          StackLinesFreeW(hStack);
        }
        else
        {
          StackFree(hStack);
          SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
        }
      }
    }
    else MessageBeep(MB_OK);
  }
}

int ReplaceSelFromStackA(HWND hWnd, HSTACK *hStack)
{
  EDITSTREAM es;
  LINEITEMSTACKA *lpElement=(LINEITEMSTACKA *)hStack->first;

  es.dwCookie=(DWORD)&lpElement;
  es.pfnCallback=ReplaceSelFromStackCallbackA;
  return SendMessage(hWnd, EM_STREAMIN, SF_TEXT|SF_UNICODE|SFF_SELECTION, (LPARAM)&es);
}

int ReplaceSelFromStackW(HWND hWnd, HSTACK *hStack)
{
  EDITSTREAM es;
  LINEITEMSTACKW *lpElement=(LINEITEMSTACKW *)hStack->first;

  es.dwCookie=(DWORD)&lpElement;
  es.pfnCallback=ReplaceSelFromStackCallbackW;
  return SendMessage(hWnd, EM_STREAMIN, SF_TEXT|SF_UNICODE|SFF_SELECTION, (LPARAM)&es);
}

DWORD CALLBACK ReplaceSelFromStackCallbackA(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
  LINEITEMSTACKA **lpElement=(LINEITEMSTACKA **)dwCookie;
  wchar_t *wpDest=(wchar_t *)pbBuff;
  int nDestSize=cb / sizeof(wchar_t) - 2;
  int nDestCount=0;
  static int nSrcCount;

  if (lpElement && *lpElement)
  {
    MultiByteToWideChar(CP_ACP, 0, (*lpElement)->pLine + nSrcCount, (*lpElement)->nLineLen - nSrcCount, wpDest, nDestSize);
    nSrcCount+=nDestCount=min((*lpElement)->nLineLen - nSrcCount, nDestSize);

    if (nSrcCount >= (*lpElement)->nLineLen)
    {
      wpDest[nDestCount++]='\n';
      *lpElement=(*lpElement)->next;
      nSrcCount=0;
    }
    *pcb=nDestCount * sizeof(wchar_t);
  }
  else *pcb=0;

  return 0;
}

DWORD CALLBACK ReplaceSelFromStackCallbackW(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
  LINEITEMSTACKW **lpElement=(LINEITEMSTACKW **)dwCookie;
  wchar_t *wpDest=(wchar_t *)pbBuff;
  int nDestSize=cb / sizeof(wchar_t) - 2;
  int nDestCount=0;
  static int nSrcCount;

  if (lpElement && *lpElement)
  {
    while (nDestCount < nDestSize && nSrcCount < (*lpElement)->nLineLen)
      wpDest[nDestCount++]=(*lpElement)->wpLine[nSrcCount++];

    if (nSrcCount >= (*lpElement)->nLineLen)
    {
      wpDest[nDestCount++]='\n';
      *lpElement=(*lpElement)->next;
      nSrcCount=0;
    }
    *pcb=nDestCount * sizeof(wchar_t);
  }
  else *pcb=0;

  return 0;
}

DUPLICATEITEMA* StackPushRemoveDuplicatesA(HSTACK *hStack, char *pLine, int nLineLen, char *pLineStr, int nLineStrLen)
{
  DUPLICATEITEMA *lpElement=(DUPLICATEITEMA *)hStack->first;
  DUPLICATEITEMA *lpNewElement=NULL;

  while (lpElement)
  {
    if (lpElement->nLineStrLen == nLineStrLen && !xstrcmpinA(lpElement->pLineStr, pLineStr, nLineStrLen))
      return NULL;

    lpElement=lpElement->next;
  }
  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpNewElement, -1, sizeof(DUPLICATEITEMA)))
  {
    lpNewElement->pLine=pLine;
    lpNewElement->nLineLen=nLineLen;
    lpNewElement->pLineStr=pLineStr;
    lpNewElement->nLineStrLen=nLineStrLen;
  }
  return lpNewElement;
}

DUPLICATEITEMW* StackPushRemoveDuplicatesW(HSTACK *hStack, wchar_t *wpLine, int nLineLen, wchar_t *wpLineStr, int nLineStrLen)
{
  DUPLICATEITEMW *lpElement=(DUPLICATEITEMW *)hStack->first;
  DUPLICATEITEMW *lpNewElement=NULL;

  while (lpElement)
  {
    if (lpElement->nLineStrLen == nLineStrLen && !xstrcmpinW(lpElement->wpLineStr, wpLineStr, nLineStrLen))
      return NULL;

    lpElement=lpElement->next;
  }
  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpNewElement, -1, sizeof(DUPLICATEITEMW)))
  {
    lpNewElement->wpLine=wpLine;
    lpNewElement->nLineLen=nLineLen;
    lpNewElement->wpLineStr=wpLineStr;
    lpNewElement->nLineStrLen=nLineStrLen;
  }
  return lpNewElement;
}

LINEITEMSTACKA* StackInsertLineA(HSTACK *hStack, char *pLine, int nLineLen, int nIndex)
{
  LINEITEMSTACKA *lpElement=NULL;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, nIndex, sizeof(LINEITEMSTACKA)))
  {
    lpElement->pLine=pLine;
    lpElement->nLineLen=nLineLen;
  }
  return lpElement;
}

LINEITEMSTACKW* StackInsertLineW(HSTACK *hStack, wchar_t *wpLine, int nLineLen, int nIndex)
{
  LINEITEMSTACKW *lpElement=NULL;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, nIndex, sizeof(LINEITEMSTACKW)))
  {
    lpElement->wpLine=wpLine;
    lpElement->nLineLen=nLineLen;
  }
  return lpElement;
}

void StackLinesFreeA(HSTACK *hStack)
{
  LINEITEMSTACKA *lpElement=(LINEITEMSTACKA *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->pLine);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackLinesFreeW(HSTACK *hStack)
{
  LINEITEMSTACKW *lpElement=(LINEITEMSTACKW *)hStack->first;

  while (lpElement)
  {
    GlobalFree((HGLOBAL)lpElement->wpLine);

    lpElement=lpElement->next;
  }
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

void StackFree(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

int CompareLineStrA(const void *elem1, const void *elem2)
{
  LINEITEMARRAYA *lpItem1=(LINEITEMARRAYA *)elem1;
  LINEITEMARRAYA *lpItem2=(LINEITEMARRAYA *)elem2;
  int nResult;

  nResult=xstrcmpiA(lpItem1->pLine + lpItem1->nOffset, lpItem2->pLine + lpItem2->nOffset);
  if (!nResult && (lpItem1->nOffset || lpItem2->nOffset))
    nResult=xstrcmpiA(lpItem1->pLine, lpItem2->pLine);
  return nResult;
}

int CompareLineStrW(const void *elem1, const void *elem2)
{
  LINEITEMARRAYW *lpItem1=(LINEITEMARRAYW *)elem1;
  LINEITEMARRAYW *lpItem2=(LINEITEMARRAYW *)elem2;
  int nResult;

  nResult=xstrcmpiW(lpItem1->wpLine + lpItem1->nOffset, lpItem2->wpLine + lpItem2->nOffset);
  if (!nResult && (lpItem1->nOffset || lpItem2->nOffset))
    nResult=xstrcmpiW(lpItem1->wpLine, lpItem2->wpLine);
  return nResult;
}

int CompareLineIntA(const void *elem1, const void *elem2)
{
  LINEITEMARRAYA *lpItem1=(LINEITEMARRAYA *)elem1;
  LINEITEMARRAYA *lpItem2=(LINEITEMARRAYA *)elem2;
  char *pLine1=lpItem1->pLine + lpItem1->nOffset;
  char *pLine2=lpItem2->pLine + lpItem2->nOffset;
  int nLineInt1;
  int nLineInt2;

  while (*pLine1 && (*pLine1 < '0' || *pLine1 > '9'))
    ++pLine1;
  nLineInt1=xatoiA(pLine1);

  while (*pLine2 && (*pLine2 < '0' || *pLine2 > '9'))
    ++pLine2;
  nLineInt2=xatoiA(pLine2);

  if (nLineInt1 == nLineInt2)
    return CompareLineStrA(elem1, elem2);
  else
    return nLineInt1 - nLineInt2;
}

int CompareLineIntW(const void *elem1, const void *elem2)
{
  LINEITEMARRAYW *lpItem1=(LINEITEMARRAYW *)elem1;
  LINEITEMARRAYW *lpItem2=(LINEITEMARRAYW *)elem2;
  wchar_t *wpLine1=lpItem1->wpLine + lpItem1->nOffset;
  wchar_t *wpLine2=lpItem2->wpLine + lpItem2->nOffset;
  int nLineInt1;
  int nLineInt2;

  while (*wpLine1 && (*wpLine1 < '0' || *wpLine1 > '9'))
    ++wpLine1;
  nLineInt1=xatoiW(wpLine1);

  while (*wpLine2 && (*wpLine2 < '0' || *wpLine2 > '9'))
    ++wpLine2;
  nLineInt2=xatoiW(wpLine2);

  if (nLineInt1 == nLineInt2)
    return CompareLineStrW(elem1, elem2);
  else
    return nLineInt1 - nLineInt2;
}

void swap_item(register unsigned char *src, register unsigned char *dest, register unsigned int size)
{
  register int temp;

  if (src != dest)
  {
    while (size--)
    {
      temp=*src;
      *src++=*dest;
      *dest++=temp;
    }
  }
}

void quicksort(void *base, unsigned int count, unsigned int width, int (*compare)(const void *, const void *))
{
  unsigned char *low;
  unsigned char *high;
  unsigned char *middle;
  unsigned char *lowguy;
  unsigned char *highguy;
  unsigned char *lowstk[30];
  unsigned char *highstk[30];
  int size;
  int stkptr;

  if (count < 2 || width == 0)
    return;

  stkptr=0;
  low=(unsigned char *)base;
  high=(unsigned char *)base + width * (count - 1);

  recurse:

  size=(high - low) / width + 1;
  middle=low + (size / 2) * width;
  swap_item(middle, low, width);
  lowguy=low;
  highguy=high + width;

  for (;;)
  {
    do
    {
      lowguy+=width;
    }
    while (lowguy <= high && compare(lowguy, low) < 0);

    do
    {
      highguy-=width;
    }
    while (highguy > low && compare(highguy, low) > 0);

    if (highguy < lowguy)
      break;

    swap_item(lowguy, highguy, width);
  }

  swap_item(low, highguy, width);

  if (highguy - 1 - low >= high - lowguy)
  {
    if (low + width < highguy)
    {
      lowstk[stkptr]=low;
      highstk[stkptr]=highguy - width;
      ++stkptr;
    }
    if (lowguy < high)
    {
      low=lowguy;
      goto recurse;
    }
  }
  else
  {
    if (lowguy < high)
    {
      lowstk[stkptr]=lowguy;
      highstk[stkptr]=high;
      ++stkptr;
    }
    if (low + width < highguy)
    {
      high=highguy - width;
      goto recurse;
    }
  }

  if (--stkptr >= 0)
  {
    low=lowstk[stkptr];
    high=highstk[stkptr];
    goto recurse;
  }
  else return;
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_WORDWRAPOFF)
      return L"\x041F\x0435\x0440\x0435\x043D\x043E\x0441\x0020\x043F\x043E\x0020\x0441\x043B\x043E\x0432\x0430\x043C\x0020\x0434\x043E\x043B\x0436\x0435\x043D\x0020\x0431\x044B\x0442\x044C\x0020\x0432\x043A\x043B\x044E\x0447\x0435\x043D\x002E";
  }
  else
  {
    if (nStringID == STRID_WORDWRAPOFF)
      return L"Word wrap must be turned on.";
  }
  return L"";
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hMainWnd=pd->hMainWnd;
  hWndEdit=pd->hWndEdit;
  bOldWindows=pd->bOldWindows;
  bAkelEdit=pd->bAkelEdit;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
  }
  return TRUE;
}
